"""
Unit tests.
"""

from heliokos.domain.core import Concept, ConceptRepo, SKOS, Literal


def test_narrower():
    physics = Concept()
    physics.pref_label = "Physics"
    plasma = Concept()
    plasma.pref_label = "Plasma"
    physics.add_narrower(plasma)
    repo = ConceptRepo()
    repo.add(physics)
    repo.add(plasma)
    assert repo.g.value(physics.id, SKOS.narrower) == plasma.id


def test_concept_repo():
    repo = ConceptRepo()
    solar_wind = Concept()
    solar_wind.pref_label = "Solar wind"
    repo.add(solar_wind)
    assert any(repo.g.triples((None, SKOS.prefLabel, Literal("Solar wind"))))



"""

- one HTTP Route, returning HTML Representation of Resource

infra:
- FastAPI app returning HTML responses
- Jinja2 to render Representation of Resource
"""

from fastapi import FastAPI
from starlette import status
from starlette.requests import Request
from starlette.responses import RedirectResponse, HTMLResponse

from heliokos.domain.core import core_repo, Concept
from heliokos.ui.html import page_for

app = FastAPI()


@app.get("/", include_in_schema=False)
def read_root(req: Request):
    return RedirectResponse(req.url_for("read_resource", id_="home"))


@app.get("/resources/{id_}")
def read_resource(id_: str, req: Request):
    print(id_)
    base = "https://n2t.net/ark:57802/p03295/"
    if ":" in id_:
        base_nickname, id_ = id_.split(":", maxsplit=1)

        if base_nickname == "rdf":
            id_ = "http://www.w3.org/1999/02/22-rdf-syntax-ns#" + id_

    concept = Concept({"@id": id_, "@context": {"@base": base}})
    print(concept)
    inbound, outbound = core_repo.neighborhood_for(concept)
    if len(outbound) == 0:
        return HTMLResponse(f"{id_} not found", status_code=status.HTTP_404_NOT_FOUND)
    return HTMLResponse(
        content=page_for(concept=concept, inbound=inbound, outbound=outbound),
    )



from pathlib import Path

from rdflib.namespace import SKOS
from rdflib import Literal

from heliokos.infra.core import (
    RDFGraphRepo,
    RDFGraphDocument,
)


class Concept(RDFGraphDocument):
    """
    Represent a skos:Concept.

    Support setting skos:narrower only, with skos:broader inferrable.
    Rationale: https://www.w3.org/TR/vocab-data-cube/#schemes-hierarchy
    """

    def __init__(self, init_data=None):
        super().__init__(init_data)

    @property
    def pref_label(self):
        return self.g.value(self.id, SKOS.prefLabel)

    @pref_label.setter
    def pref_label(self, s):
        self.g.add((self.id, SKOS.prefLabel, Literal(s)))

    @property
    def alt_labels(self):
        return list(self.g.objects(self.id, SKOS.altLabel))

    def add_alt_label(self, s):
        self.g.add((self.id, SKOS.altLabel, Literal(s)))

    @property
    def narrower(self):
        return list(self.g.objects(self.id, SKOS.narrower))

    def add_narrower(self, concept: "Concept"):
        self.g.add((self.id, SKOS.narrower, concept.id))


class ConceptRepo(RDFGraphRepo):
    def __init__(self):
        super().__init__()


core_repo = RDFGraphRepo()
for ttl_file in Path(__file__).parent.glob("*.ttl"):
    core_repo.add_from_file(ttl_file)
for ttl_file in Path(__file__).parent.parent.joinpath("infra").glob("rd*.ttl"):
    core_repo.add_from_file(ttl_file)



import json
import uuid
from pathlib import Path

from rdflib import Graph, URIRef
from rdflib.namespace import SKOS, RDFS
from toolz import merge

RDFA_CORE_INITIAL_CONTEXT = json.loads(
    Path(__file__).parent.joinpath("rdfa11.json").read_text()
)


def with_core_context(d):
    return merge(d, RDFA_CORE_INITIAL_CONTEXT)


def with_uuid_id(d):
    return merge(d, {"@id": str(uuid.uuid4())})


class ValidationError(Exception):
    pass


class RDFGraphDocument:
    """
    A rdflib graph fragment where there is one and only one "root" subject.
    """

    def __init__(self, init_data=None):
        self.g = Graph()
        data = {} if init_data is None else init_data
        if len(data) == 0:
            data = with_uuid_id(with_core_context({"@type": "rdfs:Resource"}))
        if "@context" not in data:
            data = with_core_context(data)
        if "@id" not in data:
            data = with_uuid_id(data)
        if "@type" not in data:
            data["@type"] = "rdfs:Resource"
        self.g.parse(
            data=data,
            format="json-ld",
        )
        self.id = next(self.g.subjects())
        self.check()

    @property
    def label(self):
        rv = self.g.value(self.id, SKOS.prefLabel)
        if rv is None:
            rv = self.g.value(self.id, RDFS.label)
        if rv is None:
            rv = f"@id={self.id}"
        return rv

    def update(self, doc):
        self.g.parse(data=doc, format="json-ld")
        self.check()

    def check(self):
        if len(list(self.g.subjects(unique=True))) != 1:
            raise ValidationError(f"document '{self.label}' has more than one subject")

    def __repr__(self):
        return json.dumps(json.loads(self.g.serialize(format="json-ld"))[0], indent=2)


class RDFGraphRepo:
    def __init__(self):
        self.g = Graph()

    def add(self, doc: RDFGraphDocument):
        for triple in doc.g:
            self.g.add(triple)

    def add_from_file(self, path: Path):
        self.g.parse(path)

    def get_document_by_id(self, id_):
        doc_g = Graph()
        for t in self.g.triples((URIRef(id_), None, None)):
            doc_g.add(t)
        return json.loads(doc_g.serialize(format="json-ld"))[0]

    def neighborhood_for(self, concept):
        outbound = [t for t in self.g.triples((concept.id, None, None))]
        inbound = [t for t in self.g.triples((None, None, concept.id))]
        return inbound, outbound



from jinja2 import Environment, PackageLoader, select_autoescape
from rdflib import SKOS

from heliokos.domain.core import core_repo

jinja_env = Environment(
    loader=PackageLoader("heliokos.ui", "templates"),
    autoescape=select_autoescape(),
)

REGION_NS = "https://n2t.net/ark:57802/p03295/"


def page_for(concept, inbound, outbound):
    concept_link = concept.id.removeprefix(REGION_NS)
    inbound_local = [
        (s.removeprefix(REGION_NS), p, o.removeprefix(REGION_NS)) for s, p, o in inbound
    ]
    outbound_local = [
        (s.removeprefix(REGION_NS), p, o.removeprefix(REGION_NS))
        for s, p, o in outbound
    ]
    concept_preflabel = core_repo.g.value(concept.id, SKOS.prefLabel)
    return jinja_env.get_template("concept_neighborhood.html").render(
        concept_preflabel=concept_preflabel,
        concept_link=concept_link,
        inbound=inbound_local,
        outbound=outbound_local,
    )
